// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef DYN_MODULES_AUTOMATON_HH
#define DYN_MODULES_AUTOMATON_HH

#include <unordered_map>
#include <awali/dyn/core/automaton.hh>
#include <awali/dyn/loading/handler.hh>
#include <awali/dyn/core/context-description.hh>

namespace awali {
  namespace dyn {
    
    /** Makes a deep copy of given automaton.
     *  
     *  @param aut Automaton to copy
     *  @return a copy of aut
     */
    automaton_t copy(automaton_t aut);

    /** Computes the support of given automaton, that is the boolean automaton resulting from replacing any non-zero weight by True.
     *  
     * @param aut automaton the support of which is computed
     * @return automaton over B
     * */
    automaton_t support(automaton_t aut);

    /** Computes the characteristic of `aut` in the given semi-ring,
     * that is setting the weight of every transition to one (i.e. the
     * multiplicative neutral element of the weightset).
     *  
     *  @param aut 
     *  @param semiring string representation of a semiring
     *
     *  @pre `aut` must be a boolean automaton
     */
    automaton_t characteristic(automaton_t aut, std::string semiring);

    /** Determines if two automata are isomorphic.
     *  
     *  @param aut
     *  @param aut
     *  @pre the two automata must have the same context
     */
    bool are_isomorphic(automaton_t aut1, automaton_t aut2);
    
    /** In a copy of given automaton, multiplies the weight of each
     * transition by a constant on the left.
     * (No difference with {@link right_mult} if the weightset of `aut` is
     * commutative).
     * @param aut automaton (unchanged)
     * @param w weight by which the transition weight are multiplied
     * @return a copy of `aut` in which the transition weights have been
     *         multiplied by `w`
     * @see left_mult_here
     */
    automaton_t left_mult(automaton_t aut, weight_t w);
    
    /** Multiplies the weight of each transition by a constant on the
     * left.
     * (No difference with {@link right_mult_here} if the weightset of `aut` is
     * commutative).
     * @param aut (modified)
     * @param w weight by which the transition weight are multiplied
     * @see left_mult
     */
    void left_mult_here(automaton_t aut, weight_t w);

    /** In a copy of given automaton, multiplies the weight of each
     * transition by a constant on the right.
     * (No difference with {@link left_mult} if the weightset of `aut` is
     * commutative).
     * @param aut automaton (unchanged)
     * @param w weight by which the transition weight are multiplied
     * @return a copy of `aut` in which the transition weights have been
     *         multiplied by `w`
     * @see right_mult_here
     */
    automaton_t right_mult(automaton_t aut, weight_t w);


    /** Multiplies the weight of each transition by a constant on the
     * right.
     * (No difference with {@link left_mult_here} if the weightset of `aut` is
     * commutative).
     * @param aut (modified)
     * @param w weight by which the transition weight are multiplied
     * @see right_mult
     */
    void right_mult_here(automaton_t aut, weight_t w);

  }
}//end of ns awali::dyn

#endif
