// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef DYN_MODULES_CONTEXT_CC
#define DYN_MODULES_CONTEXT_CC

#include <set>

#include <awali/dyn/modules/context.hh>
#include <awali/dyn/loading/handler.hh>


namespace awali { namespace dyn {

  automaton_t make_automaton_from_context(context_t ctx) {
    return loading::call0<automaton_t>("make_automaton_from_context", "context", ctx->sname(), ctx);
  }

  automaton_t make_automaton_from_context(context::context_description cd) {
    std::string stat_ctx = context::tostring(cd, false);
      return loading::call0<automaton_t>("make_automaton_from_desc", "context", stat_ctx, cd);
  }

  automaton_t make_automaton_from_context(context::labelset_description ls, context::weightset_description ws) {
    return make_automaton_from_context(c_desc(ls, ws));
  }

  context_t make_context(context::context_description cd) {
    return loading::call0<context_t>("make_context_from_desc", "context", tostring(cd,false), cd);
  }

  automaton_t make_automaton(std::string alphabet, std::string semiring) {
    return make_automaton_from_context(context::letterset(alphabet),context::weightset(semiring));
  }

    automaton_t make_int_automaton(int a, int b, std::string semiring) {
      return make_automaton_from_context(context::intletterset(a,b),context::weightset(semiring));
    }

    automaton_t make_int_automaton(int n, std::string semiring) {
      return make_automaton_from_context(context::intletterset(n),context::weightset(semiring));
    }

  automaton_t make_automaton(std::string alphabet) {
    return make_automaton(alphabet, "B");
  }

  automaton_t make_automaton_with_eps(std::string alphabet, std::string semiring) {
    return make_automaton_from_context(nullableset(context::letterset(alphabet)),context::weightset(semiring));
  }

    automaton_t make_int_automaton_with_eps(int a, int b, std::string semiring) {
      return make_automaton_from_context(nullableset(context::intletterset(a,b)),context::weightset(semiring));
    }

    automaton_t make_int_automaton_with_eps(int n, std::string semiring) {
      return make_automaton_from_context(nullableset(context::intletterset(n)),context::weightset(semiring));
    }

  automaton_t make_automaton_with_eps(std::string alphabet) {
    return make_automaton_with_eps(alphabet, "B");
  }
    
  namespace internal {
    automaton_t parse_automaton(js_object* p) {
      if(!p->has_field("kind") || p->str("kind")!="Automaton")
        throw std::runtime_error("json: Automaton");
      if(!p->has_field("context"))
        throw std::runtime_error("No context in Automaton description");
      dyn::context::context_description ct= dyn::context::parse_context(p->object("context"));
      std::string stat_ctx = tostring(ct,false);
      typedef automaton_t (*bridge_t)(js_object*, dyn::context::context_description);
      static std::unordered_map<std::string, void*> bridges;
      auto it = bridges.find(stat_ctx);
      automaton_t res;
      if(it == bridges.end()) {
        auto bridge = (bridge_t) loading::get_handler("parse_automaton", "context", stat_ctx);
        bridges.emplace(stat_ctx, (void*) bridge);
        res= bridge(p, ct);
      }
      else
        res = ((bridge_t) it->second)(p, ct);
      return res;
    }

    automaton_t parse_automaton(std::istream& i) {
      js_object* p=js_object::parse(i);
      auto res=internal::parse_automaton(p);
      delete p;
      return res;
    }
  }
    

  ratexp_t make_ratexp_with_context(const std::string& exp, context::context_description cd) {
    std::string stat_ctx = tostring(cd, false);
    typedef ratexp_t (*bridge_t)(const char*, context::context_description);
    static std::unordered_map<std::string, void*> bridges;
    auto it = bridges.find(stat_ctx);
    if(it == bridges.end()) {
      auto bridge = (bridge_t) loading::get_handler("make_ratexp_with_desc", "context", stat_ctx);
      bridges.emplace(stat_ctx, (void*) bridge);
      return bridge(exp.c_str(), cd);
    }
    else
      return ((bridge_t) it->second)(exp.c_str(), cd);
  }

  ratexp_t make_ratexp_with_context(const std::string& exp, context::labelset_description ls, context::weightset_description ws) {
    return make_ratexp_with_context(exp, c_desc(ls, ws));
  }

  ratexp_t make_ratexp_with_alphabet(const std::string& exp, const std::string& alph, const std::string& sr) {
    return make_ratexp_with_context(exp, context::letterset(alph),context::weightset(sr));
  }
    
    ratexp_t make_ratexp_with_alphabets(const std::string& exp, std::vector<std::string> alphabets, std::string semiring) {
      std::vector<context::labelset_description> v;
      for(auto al : alphabets)
        v.emplace_back(context::nullableset(context::letterset(al)));
      return make_ratexp_with_context(exp, context::ltupleset(v),context::weightset(semiring));
    }
    
    ratexp_t make_ratexp_with_alphabet(const std::string& exp, const std::string& alph) {
      return make_ratexp_with_alphabet(exp, alph, "B");
    }
    
    ratexp_t make_ratexp_with_alphabets(const std::string& exp, std::vector<std::string> alphabets) {
      return make_ratexp_with_alphabets(exp, alphabets, "B");
    }
    ratexp_t make_int_ratexp_with_range(const std::string& exp, int a, int b, const std::string& sr) {
    return make_ratexp_with_context(exp, context::intletterset(a,b),context::weightset(sr));
  }

    ratexp_t make_int_ratexp_with_range(const std::string& exp, int a, int b) {
      return make_int_ratexp_with_range(exp, a, b, "B");
    }

  ratexp_t make_ratexp(const std::string& exp, const std::string& sr) {
    std::set<char> alph;
    unsigned inib=0;
    for(size_t i=0; i<exp.length(); ++i) {
      if(exp[i]=='<' || exp[i]=='{')
        ++inib;
        else if(exp[i]=='>' || exp[i]=='}')
          --inib;
        else if(inib==0)
          if(exp[i]!='(' && exp[i]!=')' && exp[i]!='.' && exp[i]!='+' && exp[i]!='*' && exp[i]!='?') {
            if ((exp[i] != '\\') || (exp[i+1] != 'e'))
              alph.emplace(exp[i]);
            else
              i++;
          }
    }
    std::string a="";
    for(char c : alph)
      a=a+c;
    return make_ratexp_with_alphabet(exp, a, sr);
  }

  ratexp_t make_ratexp(const std::string& exp) {
    return make_ratexp(exp, "B");
  }



  ratexp_t parse_ratexp(std::istream& i) {
    js_object* p = js_object::parse(i);
    if(!p->has_field("kind") || p->str("kind")!="Rational Expression")
      throw std::runtime_error("json: Rational Expression");
    dyn::context::context_description ct= dyn::context::parse_context(p->object("context"));
    std::string stat_ctx = tostring(ct,false);
    typedef ratexp_t (*bridge_t)(context::context_description, js_object*);
    static std::unordered_map<std::string, void*> bridges;
    auto it = bridges.find(stat_ctx);
    ratexp_t res;
    if(it == bridges.end()) {
      auto bridge = (bridge_t) loading::get_handler("parse_ratexp", "context", stat_ctx);
      bridges.emplace(stat_ctx, (void*) bridge);
      res= bridge(ct, p->object("data"));
    }
    else
      res = ((bridge_t)it->second)(ct, p->object("data"));
    delete p;
    return res;
  }
}}//end of ns awali::dyn

#endif
