// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef DYN_DETERMINIZE_HH
#define DYN_DETERMINIZE_HH

#include <awali/dyn/core/automaton.hh>

//This module must not be used with zmin or zmax...
//Only for lal automata

namespace awali { namespace dyn {

  /** Determinize an automaton.
   * @param aut (unchanged)
   * @param history whether history of the new automaton should be populated
   *        with sets of states of `aut`
   * @pre `aut` should be over weighset B
   * @return The derminization of `aut`
   */
  automaton_t determinize(automaton_t aut, bool history = true);

  /** Computes an automaton accepting the complement of.the language accepted by
   * a given automaton.
   * @param aut automaton to be complemented
   * @return
   * @pre `aut` should be deterministic and over weightset B
   */
  automaton_t complement(automaton_t aut);
  
  /** Modifies an automaton so that it accepts the complement of the language
   * it used to accept.
   * @aut automaton to complement
   * @pre `aut` should be deterministic and over weightset B
   */
  void complement_here(automaton_t aut);

  /** Returns the completion of a copy of given automaton. 
   * @param aut (unchanged)
   * @return a complete automaton
   */
  automaton_t complete(automaton_t aut);

  /** Modifies an automaton so that it becomes complete.
   * @param aut the automaton to complete
   * @pre the labels of `aut` should be letters (in particular without epsilon
   * transition)
   */
  void complete_here(automaton_t aut);

  //Automata over fields or z only
  
  /** Reduces an automaton.
   *
   * @pre the weightset of `aut` should be a field or Z.
   */
  automaton_t reduce(automaton_t aut);
  
  automaton_t left_reduce(automaton_t aut);

  /** Tests whether an automaton is deterministic.
   * That is, whether there is a state with two outgoing transitions labelled
   * by the same letter.
   * @pre the labels of `aut` should be letters.
   */
  bool is_deterministic(automaton_t aut);


  /** Tests whether an automaton is complete.
   * @pre the labels of `aut` should be letters.
   */
  bool is_complete(automaton_t aut);
  
  /** Tests whether an automaton is ambiguous.
   * That is, whether `aut` features two accepting runs for the same word.
   */
  bool is_ambiguous(automaton_t aut);

}}//end of ns awali::dyn

#endif
