// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef AWALI_ALGOS_JS_PARSER_HH
# define AWALI_ALGOS_JS_PARSER_HH

#include <awali/sttc/misc/raise.hh>
#include <awali/common/json.cc>
#include <awali/sttc/misc/add-epsilon-trans.hh>
#include <awali/sttc/core/rat/ratexpset.hh>
#include <awali/sttc/core/rat/ratexp.hh>
#include <awali/sttc/algos/print_exp.hh>
#include <awali/sttc/history/string_history.hh>
# include <stdexcept>
# include <stack>
# include <iostream>
# include <sstream>

namespace awali { namespace sttc {


  namespace internal {

    template<typename RatExpSet>
    struct js_exp_parser {
      using ratexpset_t = RatExpSet;
      using context_t = context_t_of<ratexpset_t>;
      using labelset_t = labelset_t_of<context_t>;
      using ratexp_t = typename ratexpset_t::value_t;
      using weightset_t = weightset_t_of<ratexpset_t>;

      js_exp_parser(const ratexpset_t& rs)
        : rs_(rs)
      {}

      ratexp_t parseNode(js_value* p) {	
	if(p->type()!=js_value::OBJECT)
	  throw std::runtime_error("js parse ratexp");
	js_object* jo=dynamic_cast<js_object*>(p);
	ratexp_t e;
	if(jo->has_field("sum")) {
	  auto& v=jo->vect("sum");
	  e=parseNode(v[0]);
	  for(unsigned i=1; i<v.size(); ++i)
	    e=rs_.add(e,parseNode(v[i]));
	}
	else if(jo->has_field("prod")) {
	  auto& v=jo->vect("prod");
	  e=parseNode(v[0]);
	  for(unsigned i=1; i<v.size(); ++i)
	    e=rs_.mul(e,parseNode(v[i]));
	}
	else if(jo->has_field("star"))
	  e= rs_.star(parseNode(jo->value("star")));
	else if(jo->has_field("label"))
	  e= rs_.atom(ls_.js_parse(jo->value("label")));
	else if(jo->has_field("one"))
	  e = rs_.one();
	else if(jo->has_field("zero"))
	  e = rs_.zero();
	else
	  throw std::runtime_error("Json parser: ratexp");

	if(jo->has_field("lweight"))
	  e=rs_.lmul(ws_.js_parse(jo->value("lweight")), e);
	if(jo->has_field("rweight"))
	  e=rs_.rmul(e, ws_.js_parse(jo->value("lweight")));
	return e;
      }
      
    private:
      const ratexpset_t& rs_;
      weightset_t ws_ = *rs_.weightset();
      labelset_t ls_ = *rs_.labelset();
      
    };
  }
    
  template <typename RatExpSet>
  inline
  typename RatExpSet::ratexp_t
  js_parse_exp_content(const RatExpSet& rs,
		       js_value* p)
  {
    internal::js_exp_parser<RatExpSet> parser{rs};
    return parser.parseNode(p);
  }

  template <typename Context>
  mutable_automaton<Context>
  js_parse_aut_content(const Context& context, js_object* p) {
    // p is a pointer on the value of the "Content" field 
    // using aut_t = mutable_automaton<Context>;
    auto ws = context.weightset();
    auto ls = context.labelset();
    mutable_automaton<Context> aut = make_mutable_automaton(context);
    char c;
    std::unordered_map<unsigned,state_t> states;
    unsigned s;
    if(!p->has_field("states"))
      throw std::runtime_error("json: states");
    for(js_value * jv : p->vect("states")) {
      js_object* jstate = dynamic_cast<js_object*>(jv); 
      if(!jstate->has_field("id"))
	throw std::runtime_error("json: State Id");
      s=jstate->intg("id");
      states[s]=aut->add_state();
      if(jstate->has_field("name"))
	aut->set_state_name(states[s],jstate->str("name"));
      if(jstate->has_field("history")) {
	auto history=aut->history();
	if(history->get_nature() == history_kind_t::NO_HISTORY) {
	  history=std::make_shared<string_history>();
	  aut->set_history(history);
	}
	auto& hs = dynamic_cast<sttc::string_history&>(*history);
	hs.add_state(states[s], jstate->str("history"));
      }
      if(jstate->has_field("initial"))
	aut->set_initial(states[s] , ws->js_parse(jstate->value("initial")));
      if(jstate->has_field("final"))
	aut->set_final(states[s] , ws->js_parse(jstate->value("final")));
    }
    if(p->has_field("transitions"))
      for(js_value * jv : p->vect("transitions")) {
	js_object* jtr = dynamic_cast<js_object*>(jv);
	if(!jtr->has_field("source"))
	  throw std::runtime_error("json: source");
	if(!jtr->has_field("destination"))
	  throw std::runtime_error("json: destination");
	s=jtr->intg("source");
	unsigned t=jtr->intg("destination");
	if(jtr->has_field("label")) {
	  auto l = ls->js_parse(jtr->value("label"));
	  if(jtr->has_field("weight")) {
	    auto w  = ws->js_parse(jtr->value("weight"));
	    aut->new_transition(states[s], states[t],l,w);
	  }
	  else
	    aut->new_transition(states[s], states[t],l);
	}
	else
	  if(jtr->has_field("weight")) {
	    auto w  = ws->js_parse(jtr->value("weight"));
	    new_epsilon_trans(aut, states[s], states[t],w);
	  }
	  else
	    new_epsilon_trans(aut, states[s], states[t]);
      }  
    return aut;
  }

  template <typename Aut>
  void
  js_add_metadata(Aut& aut, js_object* p) {
    if(p->has_field("name"))
      aut->set_name(p->str("name"));
    if(p->has_field("caption"))
      aut->set_desc(p->str("caption"));
  }    

}}//end of ns awali::stc

#endif // !AWALI_ALGOS_JS_PARSER_HH
