// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef AWALI_ALGOS_JS_PRINT_EXP_HH
# define AWALI_ALGOS_JS_PRINT_EXP_HH

#include <awali/sttc/core/rat/visitor.hh>
#include <awali/sttc/history/history.hh>
#include <awali/sttc/ctx/fwd.hh>
#include <awali/sttc/misc/raise.hh>
#include <awali/sttc/misc/json.hh>
#include <awali/common/tuple.hh>
# include <stack>
# include <iostream>
# include <iomanip>
# include <ctime>
# include <sstream>

namespace awali { namespace sttc {


  /*---------------------.
  | Json export(ratexp).  |
  `---------------------*/

  namespace rat
  {
    template <typename RatExpSet>
    class js_print_visitor
      : public RatExpSet::const_visitor
    {
    public:
      using ratexpset_t = RatExpSet;
      using context_t = context_t_of<ratexpset_t>;
      using labelset_t = labelset_t_of<context_t>;
      using ratexp_t = typename ratexpset_t::value_t;
      using weightset_t = weightset_t_of<ratexpset_t>;
      using super_type = typename ratexpset_t::const_visitor;
      using node_t = typename super_type::node_t;

     constexpr static const char* me() { return "js-print-exp"; }

      js_print_visitor(const ratexpset_t& rs, std::ostream& o)
        : rs_(rs), stream_(o)
      {}

      std::ostream& operator()(const ratexp_t& v)
      {
        v->accept(*this);
	return stream_;
      }

      AWALI_RAT_UNSUPPORTED(ldiv)
      AWALI_RAT_UNSUPPORTED(transposition)

      AWALI_RAT_VISIT(zero,)
      {
        stream_ << "\"zero\":null" ;
      }

      AWALI_RAT_VISIT(one,)
      {
       stream_ << "\"one\":null";
      }

      AWALI_RAT_VISIT(atom, e)
      {
        stream_ << "\"label\":" ;
	ls_.js_print(stream_, e.value());
      }

      AWALI_RAT_VISIT(sum, e)
      {
        stream_ << "\"sum\":[";
        for (unsigned i = 0, n = e.size(); i < n; ++i) {
	  const auto& v = e[i];
	  stream_ << '{';
	  v->accept(*this);
	  stream_ << ((i<n-1)?"},":"}");
	}
        stream_ << ']';
      }

      AWALI_RAT_VISIT(prod, e)
      {
        stream_ << "\"prod\":[";
        for (unsigned i = 0, n = e.size(); i < n; ++i) {
	  const auto& v = e[i];
	  stream_ << '{';
	  v->accept(*this);
	  stream_ << ((i<n-1)?"},":"}");
	}
        stream_ << ']';
      }

      AWALI_RAT_VISIT(conjunction, e)
      {
        stream_ << "\"conjunction\":[";
        for (unsigned i = 0, n = e.size(); i < n; ++i) {
	  const auto& v = e[i];
	  stream_ << '{';
	  v->accept(*this);
	  stream_ << ((i<n-1)?"},":"}");
	}
        stream_ << ']';
     }

      AWALI_RAT_VISIT(shuffle, e)
      {
        stream_ << "\"shuffle\":[";
        for (unsigned i = 0, n = e.size(); i < n; ++i) {
	  const auto& v = e[i];
	  stream_ << '{';
	  v->accept(*this);
	  stream_ << ((i<n-1)?"},":"}");
	}
        stream_ << ']';
      }

      AWALI_RAT_VISIT(complement, e)
      {
        stream_ << "\"node\":\"Complement\", \"value\":";
        stream_ << '{';
       e.sub()->accept(*this);
	stream_ << '}';
      }

      AWALI_RAT_VISIT(star, e)
      {
        stream_ << "\"star\":";
        stream_ << '{';
       e.sub()->accept(*this);
	stream_ << '}';
      }

      AWALI_RAT_VISIT(lweight, e)
      {
        stream_ << "\"lweight\":\"";
	ws_.print(e.weight(), stream_);
	stream_ << "\", ";
        e.sub()->accept(*this);
      }

      AWALI_RAT_VISIT(rweight, e)
      {
        stream_ << "\"rweight\":\"";
	ws_.print(e.weight(), stream_);
	stream_ << "\", ";
        e.sub()->accept(*this);
     }

    private:
      ratexpset_t rs_;
      std::ostream& stream_;
      /// Shorthand to the weightset.
      weightset_t ws_ = *rs_.weightset();
      labelset_t ls_ = *rs_.labelset();
    };

  } // rat::

  template <typename RatExpSet>
  inline
  std::ostream&
  js_print(const RatExpSet& rs,
             const typename RatExpSet::ratexp_t& e,
	   std::ostream& org)
  {
    std::ostringstream o;
    o << std::boolalpha;
    o << "{\"kind\":\"Rational Expression\",\n\"context\":";
    rs.context().js_print(o) << ',' << std::endl;
    o << "\"data\":{";
    rat::js_print_visitor<RatExpSet> printer{rs,o};
    printer(e);
    o << "}";
    std::istringstream i(o.str());
    js_formate(org, i);
    return org;
  }

  template <typename Automaton>
  inline
  std::ostream&
  js_print(Automaton aut,
	   std::ostream& org, bool full=false)
  {
    std::ostringstream o;
    auto ws=aut->context().weightset();
    auto ls=aut->context().labelset();
    o << std::boolalpha;
    o << "{\"kind\":\"Automaton\"," << std::endl;
    time_t tt;
    std::time(&tt);
    auto gmt = std::gmtime(&tt);
    o << "\"metadata\":{\"timestamp\":{";
    o << "\"day\":\"" << std::put_time(gmt,"%F") << "\", ";
    o << "\"UTCtime\":\"" << std::put_time(gmt,"%T") << "\"}";
    if(!aut->get_name().empty()) 
      o << ",\n\"name\":\"" << aut->get_name() << '\"';
    if(!aut->get_desc().empty()) 
      o << ",\n\"caption\":\"" << aut->get_desc() << '\"';
    o<< "}," << std::endl;
    o << "\"context\":";
    aut->context().js_print(o) <<',' << std::endl;
    o << "\"data\":";
    o << "{\"states\":[";
    bool first=true;
    for(unsigned i: aut->states()) {
      if(first)
	first=false;
      else
	o << ",";
      o << "{\"id\":" << (full?i:i-2) << ", \"name\":\"";
      aut->print_state(i, o);
      o << '\"';
      if(aut->has_history(i)){
      	o<< ", \"history\":\"";
      	aut->print_state_history(i, o);
	o << '\"';
      }
      if(aut->is_initial(i)){
      	o<< ", \"initial\":";
	ws-> print(aut->get_initial_weight(i),o,"json");
      }
      if(aut->is_final(i)){
      	o<< ", \"final\":";
	ws-> print(aut->get_final_weight(i),o,"json");
      }
      o << '}';
    }
    o << "]," << std::endl;// end of state list
    o << "\"transitions\":[";
    first=true;
    for(unsigned i: aut->transitions()) {
      if(first)
	first=false;
      else
	o << ',';
      o << '{';
      if(full)
	o <<"\"id\":" << i << ", ";
      o	<< "\"source\":" << (full?aut->src_of(i):aut->src_of(i)-2)
	<< ", \"destination\":" << (full?aut->dst_of(i):aut->dst_of(i)-2)
	<< ", \"label\":";
      ls->js_print(o, aut->label_of(i));
      if(!ws->is_one(aut->weight_of(i)) || ws->show_one()) {
	o << ", \"weight\":";
	ws-> print(aut->weight_of(i),o,"json") ;
      }
      o  << "}";
    }
    o << "]}"; //end of transition list + end of content value
    o << "}"; //end of global object
    std::istringstream i(o.str());
    js_formate(org, i);
    return org;
  }

}}//end of ns awali::stc

#endif // !AWALI_ALGOS_JS_PRINT_EXP_HH
