// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef AWALI_STTC_WAUTOMATON_HH
#define AWALI_STTC_WAUTOMATON_HH
#include <awali/sttc/weightset/b.hh>
#include <awali/sttc/ctx/lal_char.hh>
#include <awali/sttc/ctx/lan_char.hh>
#include <awali/sttc/ctx/lat.hh>
#include <awali/sttc/core/mutable-automaton.hh>
#include<vector>
#include <awali/sttc/algos/js_parser.hh>
#include <awali/sttc/algos/js_print.hh>

namespace awali {
  namespace sttc {

    /**
       Build an awali weighted automaton labeled by letters

       The weightset must admit a constructor without argument.

       @tparameter: the type of the weightset : z, c, zz<5>, ... (must be included)
       @parameter: an initializer list of letters; for instance {'x','y'}
       @ return : an empty automaton over the weightset and the alphabet
    **/
    template<typename T>
    mutable_automaton<context<ctx::lal_char,T>>
      make_automaton(const std::initializer_list<char>& letters)
    {
      return make_shared_ptr<mutable_automaton<context<ctx::lal_char,T>>>
	(context<ctx::lal_char,T>(ctx::lal_char(letters),T()));
    }

    /**
       Specialization of the template function for Boolean automata (NFA)
       @parameter: an initializer list of letters; for instance {'x','y'}
       @ return : an empty NFA over the given alphabet
     **/
    mutable_automaton<context<ctx::lal_char,b>>
      make_automaton(const std::initializer_list<char>& letters)
    {
      return make_automaton<b>(letters);
    }
    
    template<typename T>
    mutable_automaton<context<ctx::lat<ctx::lan_char,ctx::lan_char>,T>>
      make_transducer(const std::initializer_list<char>& letterA, const std::initializer_list<char>& letterB)
    {
      return make_shared_ptr<mutable_automaton<context<ctx::lat<ctx::lan_char,ctx::lan_char>,T>>>
	(context<ctx::lat<ctx::lan_char,ctx::lan_char>,T>(std::make_tuple(ctx::lan_char(letterA), ctx::lan_char(letterB)),T()));
    }

    template<typename T>
    mutable_automaton<context<ctx::lal_char,T>>
      load_automaton(std::istream& is) {
      awali::js_object* jo= awali::js_object::parse(is);
      if(!jo->has_field("kind") || jo->str("kind")!="Automaton")
	throw std::runtime_error("json: Automaton");
      awali::js_object* jl=jo->object("context")->object("labels");
      if(jl->str("labelKind") != "Letters" || jl->str("letterType")!= "Char")
	throw std::runtime_error("Not a letter automaton");
      if(jl->has_field("withEpsilon"))
	throw std::runtime_error("Epsilon transitions not supported by this function");
      auto ctx = context<ctx::lal_char,T>(ctx::lal_char({}),T());
      for(auto jv : jl->vect("alphabet"))
	const_cast<ctx::lal_char::genset_t&>(ctx.labelset()->genset()).add_letter(dynamic_cast<js_string*>(jv)->value[0]);
      auto aut = js_parse_aut_content(ctx, jo->object("data"));
      if(jo->has_field("metadata"))
	js_add_metadata(aut, jo->object("metadata"));
      delete jo;
      return aut;
    }
    
    mutable_automaton<context<ctx::lal_char,b>>
      load_automaton(std::istream& is) {
	return load_automaton<b>(is);
      }
  }
}

#endif
