// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <cora/edit.hh>
// #include <cora/bless.cc>
#include <sstream>
#include <unistd.h>

using namespace awali::dyn;

namespace awali {
namespace cora {
namespace details {

  const std::string tr_pack_t::DEFAULT = "@tr_pack_t::DEFAULT@";

  int
  error_message_ostream_t::custombuffer_t::overflow(int ch)
  {
    if (ch ==  '\n') {
      e.print_message(ss.str());
      ss.str("");
      refresh();
    }
    else
      ss << (char) ch;
    return ch;
  }

  void 
  editter_t::clear_state_data()
  {
    states= aut->states();
    state_area.set_size(states.size());
    state_data.clear();
    state_data = std::vector<tokens_t>(states.size());
  }

  tokens_t& editter_t::get_state_data(unsigned int i) {
    dyn::state_t s = states[i];
    tokens_t& got = state_data[i];
    if (got.size() == 0) {
      got = state_display_name(s);
    }
    return got;
  }

  void 
  editter_t::clear_transition_data()
  {
    transitions= aut->transitions();
    trans_area.set_size(transitions.size());
    transition_data.clear();
    transition_data = std::vector<tokens_t>(transitions.size());
  }

  tokens_t& editter_t::get_transition_data(unsigned int i) {
    dyn::transition_t tr = transitions[i];
    tokens_t& got = transition_data[i];
    if (got.size() == 0) {
//       tokens_t ss;
      state_t src = aut->src_of(tr);
      got << aut->get_state_name(src);
      got << tr_sta;
      weight_t w = aut->weight_of(tr); 
      got << format_weight_of(aut->weight_of(tr));
      tokens_t labels = label_format;
      if (is_transducer)
        for(const std::string& l: tdc.label_of(tr)) {
          labels /=  (l == ""? eps_repr : l);
        }
      else {
        if (aut->is_eps_transition(tr))
          labels /= eps_repr;
        else
          labels /= ctx->label_to_string(aut->label_of(tr));
      }
      got << labels;
      got << tr_end;
      state_t dst = aut->dst_of(tr);
      got << token_t(aut->get_state_name(dst));
//       transition_data[i]=ss;
//       return transition_data[i];
    }
    return got;
  }

  void editter_t::populate_alphabets() {
    alphabets = {};
    if (is_transducer) {
      for (const std::vector<char>& alphabet : tdc.alphabets()) {
        std::vector<std::string> alph = {};
        for (const char& c : alphabet) {
          alph.push_back(std::string(1,c));
        }
        alph.push_back("_");
        alphabets.push_back(alph);
      }
    } 
    else {
      std::vector<std::string> alph;
      for (const any_t& a : aut->alphabet()) 
        alph.push_back(
            ctx->label_to_string(a)
            );
      if (aut->allow_eps_transition())
        alph.push_back("_");
      alphabets.push_back(alph);
    }
  }

  editter_t::editter_t(automaton_t a, std::string f) 
    : aut(a), tdc(a), ss(*this)
  {
    ctx = a->get_context();
    states = aut->states();
    transitions = aut->transitions();
    is_transducer = aut->is_transducer();
    is_int_automaton = aut->is_int_automaton();
    has_epsilon = is_transducer || aut->allow_eps_transition();
    weightset_name = ctx->weightname();
    is_over_B = weightset_name == "B";
    if (f != "") {
      filename = f;
      is_saved = true;
    } else
      is_saved = false;

    populate_alphabets();
    // tips
    tip_t blank_tip {"",""};

    tips.push_back({ //MAIN_STATE
        {  "a", "add state"},
        {  "n", "rename state"},
        {  "d", "del. state"},
        {  "i", weightset_name=="B"?"toggle init.":"init. weight"},
        {  "f", weightset_name=="B"?"toggle final":"final weight"},
        {  "t", "trans. from"},
        (is_transducer?blank_tip:tip_t{"l","add letter"}),
        {  "w", "display"},
        {  "s", "save"},
        {  "q", "quit"}
        });
    tips.push_back({ //MAIN_TRANS
        {  "a", "add trans."},
        {  "e", "edit trans."},
        {  "d", "del. trans."},
        blank_tip,
        (is_transducer?blank_tip:tip_t{"l","add letter"}),
        {  "w", "display"},
        {  "s", "save"},
        {  "q", "quit"}
        });
    tips.push_back({ //ASK
        {"RET", "accept"},
        { "^X", "cancel"}
        });
    tips.push_back({
        { "RET", "select"},
        { "<->", "navigate"},
        {  "^X", "cancel"}
        }); //TRANSEDIT

    dyn::set_warning_stream(ss);
  }


  void 
  editter_t::run() 
  {
    initialize();

    while (continue_edit) {
      if (require_refresh)  
        action_refresh();
      if (editstate() == MAIN_TRANS) {
        action_main_trans();
      } else {
        action_main_state();
      } 
    }
    cleanup();
  }

  void
  editter_t::initialize() 
  {
    initscr();  // Start curses mode
    tr_sta = " %-%- ";
    tr_end = " %-%-%> ";
    ini_arr_format = "%-%s%-%> ";
    fin_arr_format = " %-%s%-%>";
    // label_format
    if (is_transducer)  
      label_format << tdc_lbl_sta;
    bool b = false;
    for (unsigned i = 0; i < alphabets.size(); i++) {
      if (b)
        label_format << tdc_lbl_sep;
      label_format << token_t("%s");
      b = true;
    }
    if (is_transducer)  
      label_format << tdc_lbl_end;
    // end label_format

    tr_format << "%s" << tr_sta;
    if (!is_over_B)
      tr_format << weight_sta << "%s" << weight_end;
    tr_format << label_format << tr_end << "%s" ;

    keypad(stdscr, TRUE); // Enable arrow keys
    curs_set(0);
    set_escdelay(100);
    noecho();
    compute_area_geometry();
    clear_transition_data();
    clear_state_data();
    action_refresh(); // Draw window
  }

  void
  editter_t::cleanup()
  {
    endwin();
  }



  void 
  show_area(area_t area) 
  {
    std::string blank = std::string(area.w,'#');
    attrset(A_NORMAL);
    for (int x = 0; x<area.h; x++)
      mvprintw(area.r+x, area.c, blank.c_str());
    mvprintw( area.r+(area.h-1)/2,
        area.c+(area.w-1)/2-4,
        "(%d,%d)(%d,%d)", area.r,area.c,area.h,area.w);
  }

  void 
  editter_t::compute_area_geometry()
  {
    screen.c = screen.r = 0;
    getmaxyx(stdscr, screen.h, screen.w);

    infos_area.r = state_area.r = trans_area.r = screen.r = 0;

    trans_area.w = (screen.w-2)/2;
    state_area.w = (screen.w-trans_area.w-1)/2;
    infos_area.w = screen.w - trans_area.w - state_area.w - 2;

    infos_area.c = screen.c;
    state_area.c = infos_area.c+infos_area.w+1;
    trans_area.c = state_area.c+ state_area.w+1;

    tips_area
    = { screen.h-2,     0,           2, screen.w };
    query_area
    = { tips_area.r-1,  0,           1, (screen.w*2-1)/3};
    comment_area 
    = { query_area.r-1, 0,           1, query_area.w}; 
    message_area 
    = { comment_area.r,   comment_area.w+1, 2, screen.w-comment_area.w-1};

    infos_area.h = state_area.h = trans_area.h = comment_area.r-infos_area.r-1;
  }

  void editter_t::action_refresh() {
    clear();
    compute_area_geometry();
    attron(DIM);
    for (int x = 0; x< trans_area.h; x++)
      mvaddch(trans_area.r+x,trans_area.c-1,ACS_VLINE);
    for (int x = 0; x< trans_area.h; x++)
      mvaddch(state_area.r+x,state_area.c-1,ACS_VLINE);
    attroff(DIM);

    refresh_infos();
    refresh_states();
    refresh_transitions();
    refresh_tips();
    refresh();
    require_refresh=false;
  }

  token_t 
  editter_t::format_weight_of(dyn::weight_t w, font_style_t fs) 
  {
    if ((is_over_B) || (w == ctx->weight_one() && !ctx->show_one()))
      return token_t("", fs);
    token_t res(weight_format);
    res.font_style = fs;
    return (res /= ctx->weight_to_string(w));
  }

  tokens_t 
  editter_t::state_display_name(state_t s, bool b) 
  {
    std::string name = aut->get_state_name(s);
    tokens_t res;
    if (aut->is_initial(s)) {
      token_t ini = ini_arr_format;
      res << (ini /= format_weight_of(aut->get_initial_weight(s)));
    } else
      res << "";
    res << token_t(aut->get_state_name(s), b?BOLD:NO_STYLE);
    if (aut->is_final(s)) {
      token_t fin = fin_arr_format;
      res << (fin /= format_weight_of(aut->get_final_weight(s)));
    } else
      res << "";
    return res;
  }



  void 
  editter_t::refresh_message() 
  {
    std::chrono::milliseconds ms = 
      std::chrono::duration_cast<std::chrono::milliseconds>(
          std::chrono::high_resolution_clock::now()
          - error_message_time
          );
    if (ms.count() >= error_message_timeout)
      message_area.erase();
  }

  void 
  editter_t::refresh_states(bool clear_data)
  {
    if (clear_data)
      clear_state_data();
    state_area.erase();
    state_area.print_title("States", editstate() == MAIN_STATE);
    if (states.size() == 0)
      return;
    long src = (transitions.size() >0)?aut->src_of(current_transition()):-1;
    long dst = (transitions.size() >0)?aut->dst_of(current_transition()):-1;
  
    unsigned size = 0;
    for(int i = state_area.start; i <= state_area.end(); i++)
      size += get_state_data(i).at(1).size();
    size /= (state_area.end()-state_area.start+1);

    state_area.set_size(states.size());
    for (int i = state_area.start; i <= state_area.end(); i++){
      state_t s = states[i];
      bool b = ((long)s == src) || ((long)s == dst);
      tokens_t &tokens 
        = get_state_data(i);
      if ((editstate() == MAIN_TRANS) && b)
        tokens.at(1).font_style = BOLD;
       
      std::pair<int,int> offsets 
      = tokens.align_before(1,(state_area.w-2)/2-(size+1)/2, state_area.w-2);
      state_area.print_row(
          editstate() == MAIN_STATE,
          i,
          tokens,
          offsets.first,
          offsets.second
          );
      tokens.at(1).font_style = NO_STYLE;
    }
  }

  void 
  editter_t::refresh_transitions(bool clear_data)
  {
    if (clear_data)
      clear_transition_data();
    trans_area.erase();
    trans_area.print_title("Transitions", editstate() == MAIN_TRANS);
    for (int i = trans_area.start; i <= trans_area.end(); i++){
      tokens_t& ss = get_transition_data(i);
      if (editstate() == MAIN_STATE) {
        if (aut->src_of(transitions[i]) == current_state())
          ss.at(0).font_style = BOLD;
        if (aut->dst_of(transitions[i]) == current_state())
          ss.at(ss.size()-1).font_style = BOLD;
      }
      std::pair<int,int> offsets = 
        ss.align_before(3,(trans_area.w-2)/2-2,trans_area.w-2);
      trans_area.print_row( editstate() == MAIN_TRANS, i, ss, offsets.first,
          offsets.second);

      if (editstate() == MAIN_STATE)
        for (unsigned i : {0u, unsigned(ss.size()-1)})
          ss.at(i).font_style = NO_STYLE;
    }
  }

  void 
  editter_t::refresh_tips()
  {
    tips_t cur_tips = tips[editstate()];
    tips_area.erase();
    size_t max = cur_tips.size();
    size_t i = max;
    size_t j = 1;
    int col_width = screen.w/5;
    int x = 3 + (screen.w-col_width*5)/2+col_width*4;
    while (i-->0) {
      print_at(true,  
          tips_area.r+j, 
          x-cur_tips[i].key.size(), 
          cur_tips[i].key);
      print_at(false, tips_area.r+j, x+1, cur_tips[i].hint);
      x-= col_width;
      if ( x < 0 ) {
        x = 3 + (screen.w-col_width*5)/2+col_width*4;
        j--;
      } 
    }
  }


  std::string 
  alph_displayname(std::vector<std::string> alph) 
  {
    std::stringstream ss;
    std::string sep = "{";
    for(const std::string& str : alph) {
      ss << sep;
      ss << str;
      sep = ",";
    } 
    ss << "}";
    return ss.str();
  }

//   void 
//   editter_t::print_tokens(area_t area, 
//       int yoffset, int xoffset, std::vector<std::string> tokens,
//       std::string sep) 
//   {
//     std::string s = "";
//     for(const std::string& token : tokens) {
//       int size = token.size();
//       if (xoffset +  size + (int) s.size() >= area.w) {
//         xoffset = 0;
//         yoffset++;
//         s = "";
//       }
//       print_at(false,area.r+yoffset,area.c+xoffset,s+token);
//       xoffset += s.size()+size;
//       s = sep;
//     }
//   }

  void 
  editter_t::refresh_infos(bool force) {
    static bool init = false;
    static tokens_t title_line;
    static tokens_t weightset_line;
    static std::vector<tokens_t> label_lines;
    static tokens_t eps_line;
    if (!init || force) {
      title_line = tokens_t("",NO_STYLE," ");
      weightset_line = tokens_t("",NO_STYLE," ");
      label_lines.clear();
      eps_line = tokens_t("NB: label \"_\" represents epsilon, the empty word");
      init = true;
    }


    eps_line.at(0).font_style = BOLD;

    if (title_line.size() == 0 || force)
      title_line << token_t{"Edit automaton", BOLD} << "[Unsaved]";
    if (weightset_line.size() == 0)
      weightset_line << token_t("Weight set:", BOLD) << weightset_name;
    if (label_lines.size() == 0 || force) {
      if (is_transducer)
        label_lines.push_back (tokens_t("Labels:",BOLD));
      int n_tape = 0;
      for (const std::vector<std::string>& alph : alphabets) {
        tokens_t label("", NO_STYLE, "");
        if (is_transducer)
          label << "- tape "+std::to_string(n_tape) +": ";
        else { 
          if (is_int_automaton)
            label << token_t("Labels (int): ", BOLD);
          else 
            label << token_t("Labels: ", BOLD);
        }
        label << "{";
        for(unsigned i = 0; i< alph.size(); i++)
          label << std::string(alph[i]) + (i != alph.size()-1?",":"");
        label << "}";
        label_lines.push_back(label);
        n_tape++;
      }
    }
    size_t y = 0, x = 0;
    infos_area.erase();
    infos_area.print_left(title_line, y, x);
    infos_area.new_line(y,x,2);
    infos_area.print_left(weightset_line, y, x);
    infos_area.new_line(y,x,2);
    for (const tokens_t& line: label_lines) {
      infos_area.print_left(line,y,x);
      infos_area.new_line(y,x,1);
    }
    if (has_epsilon) {
      infos_area.new_line(y,x,1);
      infos_area.print_left(eps_line,y,x);
    }
    refresh();
  }

  input_kind_t 
  editter_t::wait_for_char(char& result) 
  { 
    int x,y;
    getyx(stdscr,y,x);
    refresh_message();
    move(y,x);
    int c=getch();
    switch (c) {
      case KEY_UP : 
        return UP;
      case KEY_DOWN : 
        return DOWN;
      case KEY_LEFT : 
        return LEFT ;
      case KEY_RIGHT : 
      case '\t':
        return RIGHT;
      case KEY_ENTER:
      case '\n':
      case '\r':
        return FORWARD;
      case KEY_BACKSPACE : // Backspace on Linux
      case 127 :           // Backspace on MacOS
        return BACKSPACE;
      case KEY_RESIZE: 
        require_refresh = true;
        clear_transition_data();
        clear_state_data();
        timeout(50);
        while (KEY_RESIZE == (c = getch()));
        if (c != ERR)
          ungetch(c);
        timeout(-1);
        return OTHER;
      case KEY_DC : //Suppr 
//         case KEY_CANCEL : 
//         case KEY_UNDO : 
        return DELETE;
      case 27: //ESC key alone
      case 24: // char ^X, ctrl+x on MacOs/Linux
        return BACK;
      default:
        if (c >= 32 && c <= 126) {
          result = c;
          return NORMAL;
        } else {
          print_message(tokens_t("Unsupported character code: %s") /= std::to_string(c));
        }
    }
    return OTHER;
  }

  answer_kind_t 
  editter_t::wait_for_string(std::string& result)
  {
    //     bool no_input = true;
    curs_set(1);
    int x,y;
    getyx(stdscr,y,x);
    char c;
    input_kind_t rm;
    bool repeat = true;
    int i = result.size();
    int j = i;
    while (repeat) {
      move(y,x);
      print(true,(" "+result+" ").c_str());
      print(false, " ");
      move(y,x+j+1);
      refresh();
      rm = wait_for_char(c);
      switch (rm) {
        case NORMAL: 
          addch(c);
          result.insert(j,1,c);
          i++;
          j++;
          break; 
        case LEFT:
          if (j> 0)
            j--;
          break;
        case RIGHT:
          if (j<i)
            j++;
          break;
        case BACKSPACE:
          if (j>0)
            j--;
        case DELETE:
          if ((i > 0) && (j < i)) {
            result.erase(j,1);
            i--;
          }
          break;
        case FORWARD:
          repeat = false;
          break;
        case BACK:
          curs_set(0);
          query_area.erase();
          return ABORT;
        default: //UP, DOWN, OTHER are ignored.
          ;
      }
    }
    curs_set(0);
    query_area.erase();
    if (result == "")
      return DEFAULT;
    else
      return INPUT;
  }


  void
  editter_t::action_main_trans() {
    enter_state(MAIN_TRANS);
    refresh_transitions();
    refresh_states();
    char c;
    switch(wait_for_char(c)) { 
      case NORMAL:
        switch(c) {
          case 'q' : action_quit(); break;
          case 's' : action_save(); break;
          case 'w' : action_display(); break;
          case 'a' : action_trans_add(); break;
          case 'd' : break;
          case 'e' : action_trans_edit(); break;
          default: print_message(tokens_t("Unknown action: %s") /= c);
        }
        if (c != 'd')
          break;
      case BACKSPACE:
      case DELETE:
        action_trans_delete(); break;
      case UP:
        trans_area.action_up();
        leave_state();
        action_main_trans();
        return;
      case LEFT:
        leave_state();
        action_main_state();
        return;
      case DOWN:
        trans_area.action_down();
        leave_state();
        action_main_trans();
        return;
      case RIGHT:
        return;
      case BACK:
      case FORWARD:
      case OTHER: ; // other char are ignored
    }
  }

  void
  editter_t::action_main_state()
  {
    enter_state(MAIN_STATE);
    refresh_states();
    refresh_transitions();
    char c;
    switch(wait_for_char(c)) { 
      case NORMAL:
        switch(c) {
          case 'q' : action_quit(); return;
          case 's' : action_save(); return;
          case 'w' : action_display(); return;
          case 'a' : action_state_add(); return;
          case 'd' : break;
          case 'n' : action_state_rename(); return;
          case 'i' : action_state_initial(); return;
          case 'f' : action_state_final(); return;
          case 't' : action_state_transfrom(); return;
          case 'l' : action_add_letter(); return;
          default: print_message(
            tokens_t("Unknown action: %s") /= c);
        }
        if (c != 'd')
          break;
      case BACKSPACE:
      case DELETE:
          action_state_delete(); return;
      case UP:  
        state_area.action_up();
        refresh_states();
        break;
      case RIGHT: leave_state(); action_main_trans(); return;
      case DOWN:
        state_area.action_down();
        refresh_states();
        break;
      case LEFT:
      case BACK:
      case FORWARD:
      case OTHER: ; // other char are ignored
    }
  }

  void editter_t::action_trans_delete()
  {
    if (!transitions.empty()) {
      aut->del_transition(current_transition());
      notify_modified();
      clear_transition_data();
    }
  }

  void 
  editter_t::apply(const tr_pack_t& tr) 
  {
    // If transition exists, we delete it.
    if (aut->has_transition(tr.id))
      aut->del_transition(tr.id);
    // Then we create the one that replaces it.
    if(is_transducer){
      std::vector<std::string> labs;
      for (const std::string& label : tr.labels) {
        labs.push_back( label == eps_repr ? "" : label );
      }
      tdc.add_transition(
          aut->get_state_by_name(tr.src), 
          aut->get_state_by_name(tr.dst), 
          labs,
          ctx->get_weight(tr.weight)     );
    }
    else {
      if ((tr.labels[0] == eps_repr) && aut->allow_eps_transition())
        aut->add_eps_transition(aut->get_state_by_name(tr.src), 
            aut->get_state_by_name(tr.dst), 
            ctx->get_weight(tr.weight)     );
      else
        aut->add_transition(aut->get_state_by_name(tr.src), 
            aut->get_state_by_name(tr.dst), 
            ctx->get_label(tr.labels[0]),
            ctx->get_weight(tr.weight)     );
    }
    notify_modified();
    clear_transition_data();
  }

  void 
  editter_t::action_trans_add()
  {
    tr_pack pack(aut);
    if(edit_tr_pack(pack))
      apply(pack);
  }

  void 
  editter_t::action_trans_edit()
  {
    if (transitions.size() == 0)
      return;
    tr_pack_t pack(aut, current_transition(),eps_repr);
    if(edit_tr_pack(pack))
      apply(pack);
  }

  void 
  editter_t::action_state_transfrom()
  {
    if (states.size() == 0)
      return;
    tr_pack pack(aut);
    pack.src = aut->get_state_name(current_state());
    if(edit_tr_pack(pack,is_over_B?2:1))
      apply(pack);
  }

  bool 
  editter_t::edit_tr_pack(tr_pack_t& tr, int off) 
  {
    enter_state(TRANSEDIT);
    comment_area.erase();
    std::string question = (tr.id == -1) ? "New trans.: ": "Edit trans.: ";
    char c;
    bool repeat = true;
    int i = off;
    int maxi = tr.labels.size()+4;

    tokens_t format;
    format << token_t(question,BOLD) << " " << tr_format << "    " << "%s";
    while (repeat) {
      tokens_t to_print = format;
      int j = i;
      to_print /= token_t(tr.src == tr_pack_t::DEFAULT?"%#":tr.src,j--==0);
      if (!is_over_B)
        to_print /= token_t(tr.weight,j==0);
      j--;
      for (const std::string& label: tr.labels)
        to_print /= token_t(
          label==tr_pack_t::DEFAULT?"%#":label,
          j--==0
        );
      to_print /= token_t(
        tr.dst == tr_pack_t::DEFAULT?"%#":tr.dst,
        j--==0
      );
      to_print /= token_t(
        " OK ",
        j--==0
      );
      size_t y = 0, x =0;
      comment_area.print_left(to_print,y,x);
      refresh();
      std::string result;
      switch(wait_for_char(c)) {
        case BACK: leave_state(); comment_area.erase();return false;
        case NORMAL: 
                   result += c;
                   //FALLTRHOUGH to FORWARD
        case FORWARD:{
          bool success = false;
          if (i == 0) {
            success = ask_state(result, "Source state", 
                tr.src == tr_pack_t::DEFAULT?"":tr.src);
            if(success)
              tr.src = result;
          } 
          else if (i == maxi-2) {
            success = ask_state(result, "Destination state",
                 tr.dst == tr_pack_t::DEFAULT?"":tr.dst);
            if (success)
              tr.dst = result;
          } 
          else if (i == 1) {
            success = ask_weight(result, "Transition weight", tr.weight);
            if (success)
              tr.weight = result;
          } 
          else if (i == maxi-1) { // select "ok"
            if (tr.src == tr_pack_t::DEFAULT)
              print_message("Please enter a source state.");
            else if (tr.dst == tr_pack_t::DEFAULT)
              print_message("Please enter a destination state.");
            else {
              bool b = false;
              int i = 0;
              for (const std::string& label : tr.labels) {
                b |= (label == tr_pack_t::DEFAULT);
                if (b)
                  print_message(
                   tokens_t("Please enter label for tape %s.") /=  i
                  );
                i++;
              }
              if (!b) {
                comment_area.erase();
                leave_state();
                return true;
              }
            }
          } 
          else { //label
            int k = i-2; 
            std::stringstream ss;
            ss << "Label for tape " << k;
            success = ask_label(result, k, ss.str(), 
                 tr.labels[k] == tr_pack_t::DEFAULT?"":tr.labels[k]);
            if (success)
              tr.labels[k] = result;
          }
          query_area.erase();
          if (!success)
            break;
        }//FALLTHROUGH to RIGHT if success
        case RIGHT: 
          if (i<maxi-1)
            i++;
          if (i == 1 && is_over_B)
            i++;
          break;
        case LEFT: 
          if (i>0)
            i--; 
          if (i == 1 && is_over_B)
            i--;
          break;
        default:;
      }
    }
    comment_area.erase();
    leave_state();
    return false;
  }



  void 
  editter_t::action_state_add()
  { 
    aut->add_state();
    notify_modified();
    refresh_states(true);
  }

  void editter_t::action_state_rename(){
    if (states.size() == 0)
      return;
    std::string old = aut->get_state_name(current_state());
    std::string result;
    if (ask_state_name(result, "New name", old))
      if (result != old)  {
        aut->set_state_name(current_state(),result);
        notify_modified();
        clear_transition_data();
      }
    clear_transition_data();
    clear_state_data();
  }

  void editter_t::action_state_delete(){
    if (states.size() == 0)
      return;
    state_t s = current_state();
    size_t n = aut->in(s).size() + aut->out(s).size();

    if (n>= 1){
      std::stringstream ss;
      ss << "Deleting state " << current_state_name() << " will also delete " 
        << n << " transitions.";
      print_at(false,comment_area.r,comment_area.c,ss.str());
      bool b;
      if (ask_confirmation(b,"Confirm deletion", true))
        if (b) {
          aut->del_state(current_state());
          notify_modified();
          clear_transition_data();
        }
    } 
    else {
      aut->del_state(current_state());
      notify_modified();
    }
    comment_area.erase();
    query_area.erase();
    refresh_states(true);
    refresh_transitions(true);
  }

  void 
  editter_t::action_state_initial()
  {   
    if (states.size() == 0)
      return;
    state_t s = current_state();
    if (weightset_name == "B") {
      if (aut->is_initial(s))
        aut->unset_initial(s);
      else
        aut->set_initial(s);
      notify_modified();
    }
    else {
      std::string result;
      std::string old = ctx->weight_to_string(aut->get_initial_weight(s));
      if (ask_weight(result, "New initial weight", old))
        if (result!= old) {
          aut->set_initial(s,ctx->get_weight(result));
          notify_modified();
        }
    }
    refresh_states(true);
  }

  bool 
  editter_t::ask_weight (std::string& result, const std::string& question, 
      const std::string& def)
  {
    class weight_validator_t : public string_validator_t {
      context_t ctx;
      public:
      bool validate(const std::string& str) const {
        try {
          ctx->get_weight(str);
          return true;
        }
        catch(...) {}
        return false;
      }
      weight_validator_t(context_t context) : ctx(context) {}
    } v(ctx);
    return ask_data(result,question,def, v, "Invalid weight: %s");
  }

  bool editter_t::ask_label (std::string& result, unsigned int i, 
      const std::string& question, const std::string& def) 
  {
    class validator_t : public string_validator_t {
      const std::vector<std::string>& alph;
      public:
      bool validate(const std::string& str) const {
        for(const std::string& l : alph) {
          if (str == l)
            return true;
        }
        return false;
      }
      validator_t(std::vector<std::string>& _alph ) 
        : alph(_alph){}
    } v(alphabets[i]);
    std::stringstream ss;
    ss << "Invalid label " ;
    if (is_transducer)
      ss << "for tape " << i;
    ss << ": %s";
    return ask_data(result,question,def,v,ss.str());
  }

  bool 
  editter_t::ask_state_name (std::string& result, const std::string& question, 
      const std::string& def)
  {
    class validator_t : public string_validator_t {
      automaton_t aut;
     public:
      bool validate(const std::string& str) const {
        return !aut->has_state(aut->get_state_by_name(str));
      }
      validator_t(automaton_t _aut) : aut(_aut){};
    } v(aut);
    return ask_data(result,question,def, v, "Name already used: %s");
  }

  bool 
  editter_t::ask_filename (std::string& result, const std::string& question, 
      const std::string& def)
  {
    class validator_t : public string_validator_t {
      public:
        bool validate(const std::string& str) const {
          int res = access(str.c_str(), W_OK);
          return (res == 0 || errno == ENOENT);
        }
        validator_t() {};
    } v;
    return ask_data(result,question,def, v, "Cannot write file: %s");
  }

  bool 
  editter_t::ask_state(std::string& result, const std::string& question, 
      const std::string& def)
  {
    class validator_t : public string_validator_t {
      automaton_t aut;
      public:
      bool validate(const std::string& str) const {
        return aut->has_state(aut->get_state_by_name(str));
      }
      validator_t(automaton_t _a) : aut(_a){};
    } v(aut);
    return ask_data(result,question,def, v, "No state named: %s");
  }


  bool 
  editter_t::ask_confirmation (bool& b, const std::string& question, 
      bool def)
  {
    class validator_t : public string_validator_t {
      public:
        bool validate(const std::string& str) const {
          std::vector<std::string> answers = 
          {"yes","y","n","no","Yes","Y","N","No","Y/n","N/y"};
          for (const std::string& answer : answers)
            if (str == answer)
              return true;
          return false;
        }
        validator_t() {};
    } v;
    std::string result;
    if (ask_data(result,question,def?"Y/n":"N/y", v, "Invalid answer: %s.")) {
      b=false;
      for (const std::string& answer :
          {"yes","y","Yes","Y","Y/n"})
      if (result == answer) {
        b=true;
        break;
      }
      return true;
    }
    return false;
  }

  void editter_t::enter_state(editstate_t state) {
    editstate_stack.push(state);
    refresh_tips();
  }

  void editter_t::leave_state() {
    editstate_stack.pop();
    refresh_tips();
  }

  bool 
  editter_t::ask_data (std::string& result, 
      const std::string& question, const std::string& default_value,
      const string_validator_t& validator, const std::string& error_msg )
  {
    enter_state(ASK);
    while (true) {
      std::stringstream ss;
      ss << question <<"?";
      if (default_value != "")
        ss << "[" << default_value << "]";
      ss << " ";
      print_at(false, 
          query_area.r, query_area.c, 
          ss.str());
      switch(wait_for_string(result)) {
        case ABORT: leave_state(); return false;
        case DEFAULT: result = default_value; 
          if (default_value != "")
            { leave_state(); return true;}
        default: 
                if (validator.validate(result)) {
                  leave_state();
                  return true;
                } 
                else
                  print_message(tokens_t(error_msg) /= result);
      }
      result = "";
    }
    leave_state();
    return true;
  }

  void 
  editter_t::action_state_final()
  {
    if (states.size() == 0)
      return;
    state_t s = current_state();
    if (weightset_name == "B") {
      notify_modified();
      if (aut->is_final(s))
        aut->unset_final(s);
      else
        aut->set_final(s);
    }
    else {
      std::string result;
      std::string old_weight = ctx->weight_to_string(aut->get_final_weight(s));
      if (ask_weight(result, "New final weight", old_weight)) {
        if (result != old_weight) {
          aut->set_final(s, ctx->get_weight(result));
          notify_modified();
        }
      }
    }
    refresh_states(true);
  }


  void 
  editter_t::action_save()
  {
    std::string result;
    if (ask_filename(result, "Save automaton as", filename)) {
      filename = result;
      std::ofstream out(filename);
      out << aut ;
      out.close();
      is_saved=true;
    }
    refresh_infos();
  }

  void editter_t::notify_modified() {
    is_saved = false;
    refresh_infos();
  }

  void 
  editter_t::action_add_letter() { 
    if (is_transducer) {
      print_message("Cannot add letters to transducers yet.");
      return;
    } 

    
    if (is_int_automaton) {
      class int_validator_t : public string_validator_t {
        public: bool validate(const std::string& str) const override {
          return (str.find_first_not_of( "0123456789" ) == std::string::npos);
        }
      };
      int_validator_t validator;
      std::string result = "";
      if ( ask_data(result, "Enter the new label", "", 
              validator, "New label should be an integer." ) ) {
        aut->get_context()->add_letter(atoi(result.c_str()));
        populate_alphabets();
        refresh_infos(true);
      }
    } 
    else {
      class char_validator_t : public string_validator_t {
        public: bool validate(const std::string& str) const override {
          return ((str.size() == 1) && (str[0] != '_'));
        }
      };
      char_validator_t validator;
      std::string result = "";
      if ( ask_data(result, "Enter the new label", "", 
              validator, "New label should be of size 1." ) ) {
        aut->get_context()->add_letter(result[0]);
        populate_alphabets();
        refresh_infos(true);
      }
    }
  }

  void 
  editter_t::action_quit() { 
    if (!is_saved) {
      bool b;
      if (ask_confirmation(b, "Save before exiting",true)) {
        if (b)
          action_save();
      } 
      else // if cancel, do not quit
        return;
    }
    continue_edit = false; 
  }

  void 
  editter_t::action_display()
  {
    pdfdisplay(aut);
  }
    

  void 
  editter_t::print_message(const std::string& str) 
  {
    print_message(tokens_t(str, STANDOUT, " "));
  }

  void
  editter_t::print_message(const tokens_t& tokens) 
  {
    tokens_t t = tokens;
    t.font_style = STANDOUT;
    message_area.erase();
    message_area.print_left(t);
    error_message_time = std::chrono::high_resolution_clock::now();
  }

  // max 40 char, no newline.
  void 
  editter_t::print_error_message(const char* format, ...)
  {
    message_area.erase();
    std::this_thread::sleep_for(std::chrono::milliseconds(50));
    va_list args;
    va_start(args, format);
    //do not use print_at since vprintf cannot be wrapped twice
    attron(A_STANDOUT);
    move(message_area.r, message_area.c);
    vw_printw(stdscr,format,args);
    va_end(args);
    attroff(A_STANDOUT);
    refresh();
  }
 
  editstate_t 
  editter_t::editstate() 
  { 
    return editstate_stack.empty()?MAIN_STATE:editstate_stack.top(); 
  }
  
  state_t 
  editter_t::current_state() 
  { 
    return states[state_area.index]; 
  }

  std::string editter_t::current_state_name() 
  { 
    return aut->get_state_name(current_state()); 
  } 

  transition_t editter_t::current_transition() 
  {
    return transitions[trans_area.index];
  }


}}}// end of namespaces awali::cora::details, awali::cora, and awali

namespace awali {
namespace cora {

  void edit(automaton_t a, std::string f) {
    details::editter_t editter(a, f);
    editter.run();
  }

}}// end of namespaces awali::cora, and awali
