// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <cora/token.hh>



namespace awali { 
namespace cora {
namespace details {

  const int token_t::ARG = -1;

  token_t::token_t(font_style_t ft) 
  : font_style(ft) 
  {}

  token_t::token_t(const int& i, font_style_t ft)
  : token_t(std::to_string(i),ft)
  {}

  token_t::token_t(const std::vector<int>& v,font_style_t ft) 
  : font_style(ft) 
  {
    for (int i : v)
      *this << i;
  }

  token_t::token_t(std::initializer_list<int> l, font_style_t ft) 
  : token_t(std::vector<int>(l),ft)
  {}

  token_t::token_t(const std::string& str, font_style_t ft) 
  : token_t(ft) 
  { (*this) << str; }

  token_t::token_t(const char* str, font_style_t ft)
  : token_t(std::string(str),ft) 
  {}
  
  token_t::token_t(const char& c, font_style_t ft)
  : token_t(ft) 
  { text.push_back(c); }

  token_t::token_t(const std::string& str, bool standout)
  : token_t(str, standout?STANDOUT:NO_STYLE)
  {}
    
  token_t& 
  token_t::operator/= (const token_t& token) 
  {
    for (unsigned i =0; i<text.size(); i++) 
      if (text[i] == token_t::ARG) {
        n_args--;
        text.erase(text.begin()+i);
        text.insert(text.begin()+i, token.text.begin(),token.text.end());
        this->font_style=token.font_style;
        break;
      }
    return *this;
  }


  token_t& 
  operator <<(token_t& token, const std::string& str) 
  {
    for (unsigned i = 0; i<str.size(); i++) {
      bool b = true;
      if (str[i] == '%') {
        b = false;
        if(str[i+1] == 's') {
          token.text.push_back(token_t::ARG);
          token.n_args++;
        } 
        else if (str[i+1] == '<')
          token.text.push_back(ACS_LARROW);
        else if (str[i+1] == '>') 
          token.text.push_back(ACS_RARROW);
        else if (str[i+1] == '^')
          token.text.push_back(ACS_UARROW);
        else if (str[i+1] == 'v')
          token.text.push_back(ACS_DARROW);
        else if (str[i+1] == '-')
          token.text.push_back(ACS_HLINE);
        else if (str[i+1] == '|')
          token.text.push_back(ACS_VLINE);
        else if (str[i+1] == '#')
          token.text.push_back(ACS_DIAMOND);
        else if (str[i+1] == '.')
          token.text.push_back(ACS_BULLET);
        else
          b = true;
      }
      if (b)
        token.text.push_back(str[i]);
      else 
        i++;
    }
    return token;
  }
    

  token_t& 
  operator<<(token_t& token, int i) 
  {
    token.text.push_back(i);
    if (i == token_t::ARG)
      token.n_args++;
    return token;
  }


  void 
  token_t::print_at(int y, int x) 
  const
  {
    move(y,x);
    this->print();
  }
  
  void 
  token_t::print() 
  const
  {
    attron(this->font_style);
    for (int i : this->text)
      if (i != token_t::ARG)
        addch(i);
    attroff(this->font_style);
 }

/* ************************************************************ CLASS token_t */

  tokens_t& 
  operator <<(tokens_t& tokens, const token_t& token) 
  {
    tokens.push_back(token);
    return tokens;
  }

  tokens_t& 
  operator <<(tokens_t& tokens, const tokens_t& other) 
  {
    for (const token_t& token: other)
      tokens << token;
    return tokens;
  }

  tokens_t& 
  operator<<(tokens_t& tokens, const std::string& str) 
  {
    return tokens << token_t(str);
  }
  
  tokens_t& 
  operator<<(tokens_t& tokens, const char* str) 
  {
    return tokens << token_t(str);
  }

  tokens_t::tokens_t(const std::string& str, font_style_t fs,
  const std::string& sep_) 
  : sep(sep_), font_style(fs)
  {
    if (str != "") {
      size_t start= 0;
      size_t end = str.find(sep_,start);
      while (end != std::string::npos) {
        push_back(token_t(str.substr(start,end-start)));
        start = end+sep_.size();
        end= str.find(sep_,start);
      }
      push_back(token_t(str.substr(start)));
    }
  }

  
  size_t 
  tokens_t::length() 
  const 
  {
    size_t res = 0;
    for (const token_t& t : *this) 
      res += t.size();
    return res;
  }


  void 
  tokens_t::print_at(int y, int x) 
  const
  { 
    attron(this->font_style);
    move(y,x);
    bool b = false;
    for (const token_t &token : *this) {
      token.print();
      if (b)
        sep.print();
      b = true;
    }
    attroff(this->font_style);
  }

  std::pair<int,int> 
  tokens_t::center(unsigned width) 
  { 
    size_t length = this->length();
    if (length > width) {
      return {0,0};
    } else {
      int left = (width-length)/2;
      int right = width - length - left;
      return {left,right};
    }
  }

  std::pair<int,int> 
  tokens_t::align_before(unsigned t_pos, unsigned c_pos, unsigned c_w) 
  {
    std::pair<int,int> res = {0,0};
    size_t left = 0;
    size_t right = 0;
    for (unsigned i = 0; i<size(); i++)
      if (i<t_pos)
        left += this->at(i).text.size();
      else
        right += this->at(i).text.size();
    if (left+right > c_w) {
      left = 0;
      size_t old = 0;
      for (unsigned i = 0; i<size(); i++) {
        size_t next = old + this->at(i).text.size();
        if (next >= (c_w-2)) {
          erase(begin()+i+1,end());
          token_t& t = at(i);
          t.text.erase(t.text.begin()+c_w-old-2,t.text.end());
          break;
        }
        old = next;
      }
      push_back(token_t("..",UNDERLINE));
      return std::pair<int,int>{0,0};
    }
    else if (left > c_pos)
      res.second = c_w-(left+right);
//       push_back(blank(c_w-(left+right)));
    else if (right > c_w-c_pos)
      res.first = c_w-(left+right);
//       insert(begin(),blank(c_w-(left+right)));
    else {
      res.first = c_pos-left;
//       push_back(blank(c_w-c_pos-right));
//       insert(begin(),blank(c_pos-left));
      res.second = c_w-c_pos-right;
    }
//     return *this;
    return res;
  }

  tokens_t& 
  tokens_t::operator/= (const token_t& rep) 
  {
    for (token_t& t : *this)
      if (t.n_args>0) {
        t /= rep;
        break;
      }
    return *this;
  }


   std::ostream& operator<< (std::ostream& o, const token_t& token) {
    o << token.n_args << ":";
    for (int i : token.text) {
      if (i == token_t::ARG)
        o << "%s";
      else if (i == (long)ACS_LARROW)
        o << "%<";
      else if (i == (long)ACS_RARROW)
        o << "%>";
      else if (i == (long)ACS_UARROW)
        o << "%^";
      else if (i == (long)ACS_DARROW)
        o << "%v";
      else if (i == (long)ACS_HLINE)
        o << "%-";
      else if (i == (long)ACS_VLINE)
        o << "%|";
      else if (i == (long)ACS_DIAMOND)
        o << "%#";
      else if (i == (long)ACS_BULLET)
        o << "%.";
      else
        o << (char) i;
    }
    return o;
   }

   std::ostream& operator<< (std::ostream& o, const tokens_t& tokens) {
    bool b = false;
    for (const token_t& t: tokens) {
      if (b)
        o << tokens.sep;
      o << t;
    }
    return o;

   }


}}} // end of namespaces awali::cora::details, awali::cora, and awali.
