// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <string>
#include <unordered_map>
#include <sstream> 
#include <initializer_list>
#include <vector>

#include <ncurses.h>

#include <cora/bless.hh>

#ifndef CORA_TOKEN_HH
#define CORA_TOKEN_HH

namespace awali { 
namespace cora {
namespace details {


  // A token is essentially a string that holds a font_style_t.
  // - each character is in fact stored as an int in order to allow usage
  //   of the special characters of ncurses.
  // - we use the special token_t::ARG to allow a replacement mechanism
  //   (operator /=)
  class token_t {
   public:
    std::vector<int> text;
    font_style_t font_style = NO_STYLE;
    static const int ARG;
    unsigned int n_args = 0;

    token_t(font_style_t ft = NO_STYLE);
    token_t(const int& i, font_style_t ft = NO_STYLE);
    token_t(const std::vector<int>& v,font_style_t ft = NO_STYLE);
    token_t(std::initializer_list<int> l, font_style_t ft = NO_STYLE);
    token_t(const std::string& str, font_style_t ft = NO_STYLE);
    token_t(const char* str, font_style_t ft = NO_STYLE);
    token_t(const std::string& str, bool standout);
    token_t(const char& c, font_style_t ft = NO_STYLE);
    
    // Replace in this token any token_t::ARG with the content of the token
    // given in argument.
    token_t& operator /= (const token_t& token);
    
    size_t size() const { return text.size(); }
    void print_at(int y, int x) const;
    void print() const;
    
    bool operator !=(token_t const& other) {
      return !(*this == other);
    }

    bool operator ==(token_t const& other) {
      if (size() != other.size())
        return false;
      for (unsigned i = 0; i< size(); i++)
        if (this->text.at(i) != other.text.at(i))
          return false;
      return true;
    }
  };

  token_t& operator <<(token_t& token, const std::string& str);
  token_t& operator <<(token_t& token, int i);

  class tokens_t : public std::vector<token_t> {
  public:
    token_t sep = "";
    font_style_t font_style = NO_STYLE;
    tokens_t(font_style_t ft) : font_style(ft) {}
    tokens_t() {}
    tokens_t(const std::string& str, font_style_t = NO_STYLE, const std::string& sep =" ");
//     tokens_t(std::initializer_list<std::string> l) ;

    std::pair<int,int> 
    center(unsigned width);
    // Computes and returns blank size to put in front and back of this so that
    // the total length of tokens is c_w and that the token at position t_pos
    // starts at position c_pos.
    // If the tokens is too long, it is trimmed !
    std::pair<int,int> align_before(unsigned t_pos, unsigned c_pos, 
        unsigned c_w);

    size_t length() const;

    tokens_t& operator /= (const token_t& rep);

    void print_at(int y, int x) const; 

    bool operator ==(tokens_t const& other) {
      if (size() != other.size())
        return false;
      for (unsigned i = 0; i< size(); i++)
        if (this->at(i) != other.at(i))
          return false;
      return true;
    }
  };


  tokens_t& operator <<(tokens_t& tokens, const token_t& token);
  tokens_t& operator <<(tokens_t& tokens, const std::string& str);
  tokens_t& operator <<(tokens_t& tokens, const tokens_t& token);
  tokens_t& operator <<(tokens_t& tokens, const char* str);


   std::ostream& operator<< (std::ostream&, const token_t& token);
   std::ostream& operator<< (std::ostream&, const tokens_t& tokens);
}}} // end of namespaces awali::cora::details, awali::cora, and awali.


#endif
